
/***********************************************************************/
/* nsock_core.c -- This contains the core engine routines for the      */
/* nsock parallel socket event library.                                */
/*                                                                     */
/***********************************************************************/
/*  The nsock parallel socket event library is (C) 2000 Insecure.Com   */
/*  LLC This library is free software; you can redistribute it and/or  */
/*  modify it under the terms of the GNU General Public License as     */
/*  published by the Free Software Foundation; Version 2.  This        */
/*  guarantees your right to use, modify, and redistribute this        */
/*  software under certain conditions.  If this license is             */
/*  unacceptable to you, Insecure.Com LLC may be willing to sell       */
/* alternative licenses (contact fyodor@insecure.org ).                */
/*                                                                     */
/*  If you received these files with a written license agreement       */
/*  stating terms other than the (GPL) terms above, then that          */
/*  alternative license agreement takes precendence over this comment. */
/*                                                                     */
/*  Source is provided to this software because we believe users have  */
/*  a right to know exactly what a program is going to do before they  */
/*  run it.  This also allows you to audit the software for security   */
/*  holes (none have been found so far).                               */
/*                                                                     */
/*  Source code also allows you to port nsock to new platforms, fix    */
/*  bugs, and add new features.  You are highly encouraged to send     */
/*  your changes to Fyodor for possible incorporation into the main    */
/*  distribution.  By sending these changes to Fyodor or one the       */
/*  insecure.org development mailing lists, it is assumed that you     */
/*  are offering Fyodor the unlimited, non-exclusive right to reuse,   */
/*  modify, and relicense the code.  If you wish to specify special    */
/*  license conditions of your contributions, please state them up     */
/*  front.                                                             */
/*                                                                     */
/*  This program is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of     */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  */
/*  General Public License for more details (                          */
/*  http://www.gnu.org/copyleft/gpl.html ).                            */
/*                                                                     */
/***********************************************************************/

/* $Id: nsock_core.c,v 1.8 2003/08/16 03:29:26 fyodor Exp $ */


#include "nsock_internal.h"
#include "gh_list.h"
#include "filespace.h"

#include <assert.h>
#if HAVE_ERRNO_H
#include <errno.h>
#endif
#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#if HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif

#include "netutils.h"

/* Msock time of day -- we update this at least once per
   nsock_loop round (and after most calls that are likely to block).  
   Other nsock files should grab this 
*/
struct timeval nsock_tod;


/* Returns -1 (and sets ms->errno if there is an error so severe that
   we might as well quit waiting and have nsock_loop() return an error */
static int wait_for_events(mspool *ms, int msec_timeout) {

  int event_msecs; /* Msecs before an event goes off */
  int combined_msecs;
  struct timeval select_tv;
  struct timeval *select_tv_p;

  assert(msec_timeout >= -1);

  if (ms->evl.events_pending == 0)
    return 0; /* No need to wait on 0 events ... */

  do {  
    if (ms->evl.next_ev.tv_sec == 0) {
      event_msecs = -1; /* None of the events specified a timeout */
    } else {
      event_msecs = MAX(0, TIMEVAL_NSEC_SUBTRACT(ms->evl.next_ev, nsock_tod));
    }

    /* We cast to unsigned because we want -1 to be very high (since it means 
       no timeout) */
    combined_msecs = MIN((unsigned) event_msecs, (unsigned) msec_timeout);
    
    // printf("wait_for_events: starting wait -- combined_msecs=%d\n", combined_msecs);
    /* Set up the timeval pointer we will give to select() */
    memset(&select_tv, 0, sizeof(select_tv));
    if (combined_msecs > 0) {
      select_tv.tv_sec = combined_msecs / 1000;
      select_tv.tv_usec = (combined_msecs % 1000) * 1000;
      select_tv_p = &select_tv;
    } else if (combined_msecs == 0) {
      /* we want the tv_sec and tv_usec to be zero -- but they already are from
	 bzero */
      select_tv_p = &select_tv;
    } else {
      assert(combined_msecs == -1);
      select_tv_p = NULL;
    }
    
    
    /* Set up the descriptors for select */
    ms->mioi.fds_results_r = ms->mioi.fds_master_r;
    ms->mioi.fds_results_w = ms->mioi.fds_master_w;
    ms->mioi.fds_results_x = ms->mioi.fds_master_x;
    
    ms->mioi.results_left = select(ms->mioi.max_sd + 1, &ms->mioi.fds_results_r, &ms->mioi.fds_results_w, &ms->mioi.fds_results_x, select_tv_p);
    gettimeofday(&nsock_tod, NULL); /* Who knows how long select sat around for */

  } while (ms->mioi.results_left == -1 && socket_errno() == EINTR);
  
  if (ms->mioi.results_left == -1 && socket_errno() != EINTR) {
    ms->errnum = socket_errno();
    return -1;
  }
 
  return 0;
}


  /* A handler function is defined for each of the main event types
     (read, write, connect, timer, etc) -- the handler is called when
     new information is available for the event.  The handler makes
     any neccessary updates to the event based on any new information
     available.  If the event becomes ready for delivery, the handler
     sets nse->event_done and fills out the relevant event fields
     (status, errnum) as applicable.  The handlers also take care of
     event type specific teardown (such as clearing socket descriptors
     from select/poll lists).  If event_done is not set, the handler
     will be called again in the case of more information or an event
     timeout */

  /* The event type handlers -- the first three arguments of each are the same:
     mspool *ms
     msevent *nse -- the event we have new info on
     enum nse_status -- The reason for the call, usually NSE_STATUS_SUCCESS 
                        (which generally means a successful I/O call or 
			NSE_STATUS_TIMEOUT or NSE_STATUS_CANCELLED

     Some of the event type handlers have other parameters, specific
     to their needs.  All the handlers can assume that the calling
     function has checked that select or poll said their descriptors
     were readable/writeable (as appropriate).

     The idea is that each handler will take care of the stuff that is 
     specific to it and the calling function will handle the stuff that
     can be generalized to dispatching/deleting/etc. all events.  But the
     calling function may use type-specific info to determine whether
     the handler should be called at all (to save CPU time).
  */


/* handle_connect_results assumes that select or poll have already
   shown the descriptor to be active */
void handle_connect_result(mspool *ms, msevent *nse, 
				  enum nse_status status)
  {
  int optval, optlen = sizeof(int);
  char buf[1024];
#ifdef LINUX
  int res = 0;
#endif
  nse->event_done = 1; /* After this function, the event is always ready for
			  delivery */

  if (status == NSE_STATUS_TIMEOUT || status == NSE_STATUS_CANCELLED) {
    nse->status = status;
  } else if (status == NSE_STATUS_SUCCESS) {
    /* First we want to determine whether the socket really is connected */
    if (getsockopt(nse->iod->sd, SOL_SOCKET, SO_ERROR, (char *) &optval, &optlen) != 0)
      optval = socket_errno(); /* Stupid Solaris */
    
    switch(optval) {
    case 0:
#ifdef LINUX
      if (!FD_ISSET(nse->iod->sd, &ms->mioi.fds_results_r)) {
	/* Linux goofiness -- We need to actually test that it is writeable */
	res = send(nse->iod->sd, "", 0, 0);
	
	if (res < 0 ) {
	  nse->status = NSE_STATUS_ERROR;
	  nse->errnum = ECONNREFUSED;
	} else {
	  nse->status = NSE_STATUS_SUCCESS;
	}
      } else {
	nse->status = NSE_STATUS_SUCCESS;
      }
#else
      nse->status = NSE_STATUS_SUCCESS;
#endif
      break;
    case ECONNREFUSED:
    case EHOSTUNREACH:
      nse->status = NSE_STATUS_ERROR;
      nse->errnum = optval;
      break;
    case ENETDOWN:
    case ENETUNREACH:
    case ENETRESET:
    case ECONNABORTED:
    case ETIMEDOUT:
    case EHOSTDOWN:
      /* These ones generally don't happen, but what the heck! */
      nse->status = NSE_STATUS_ERROR;
      nse->errnum = optval;
      break;
    default:
      snprintf(buf, sizeof(buf), "Strange connect error from %s (%d)", inet_ntoa(nse->iod->peer), optval);
      perror(buf);
      assert(0); /* I'd like for someone to report it */
      break;
    }
  } else {
    assert(0); /* Currently we only know about TIMEOUT and SUCCESS callbacks */
  }

  /* In all cases, we are now done waiting for the connection */
  FD_CLR(nse->iod->sd, &ms->mioi.fds_master_r);
  FD_CLR(nse->iod->sd, &ms->mioi.fds_master_w);
  FD_CLR(nse->iod->sd, &ms->mioi.fds_master_x);
  FD_CLR(nse->iod->sd, &ms->mioi.fds_results_r);
  FD_CLR(nse->iod->sd, &ms->mioi.fds_results_w);
  FD_CLR(nse->iod->sd, &ms->mioi.fds_results_x);

  if (ms->mioi.max_sd == nse->iod->sd)
    ms->mioi.max_sd--;

  return;
}

void handle_write_result(mspool *ms, msevent *nse, 
			       enum nse_status status)  
{
  int bytesleft;
  char *str;
  int res;

  if (status == NSE_STATUS_TIMEOUT || status == NSE_STATUS_CANCELLED) {
    nse->event_done = 1;
    nse->status = status;
  } else if (status == NSE_STATUS_SUCCESS) {
    str = FILESPACE_STR(&nse->iobuf) + nse->writeinfo.written_so_far;
    bytesleft = FILESPACE_LENGTH(&nse->iobuf) - nse->writeinfo.written_so_far;
    assert(bytesleft > 0);
    res = write(nse->iod->sd, str, bytesleft);
    if (res == bytesleft) {
      nse->event_done = 1;
      nse->status = NSE_STATUS_SUCCESS;
    } else if (res >= 0) {
      nse->writeinfo.written_so_far += res;      
    } else {
      assert(res == -1);
      if (socket_errno() != EINTR && socket_errno() != EAGAIN 
#ifndef WIN32
		  && socket_errno() != EBUSY
#endif
		  ) {
	nse->event_done = 1;
	nse->status = NSE_STATUS_ERROR;
		nse->errnum = socket_errno();
      }
    }
  }

  if (nse->event_done) {
    FD_CLR(nse->iod->sd, &ms->mioi.fds_master_w);
    FD_CLR(nse->iod->sd, &ms->mioi.fds_results_w);
    // Note -- I only want to decrement IOD if there are no other events hinging on it. 
    // For example, there could be a READ and WRITE outstanding at once
    if (nse->iod->events_pending <= 1 && ms->mioi.max_sd == nse->iod->sd)
      ms->mioi.max_sd--;
  }

  return;
}

void handle_timer_result(mspool *ms, msevent *nse, 
			       enum nse_status status)  
{

  /* Ooh this is a hard job :) */

  nse->event_done = 1;
  nse->status = status;

  return;
}

void handle_read_result(mspool *ms, msevent *nse, 
			       enum nse_status status)  
{

  char buf[4096];
  int buflen;
  unsigned int count, len;
  char *str;
  
  if (status == NSE_STATUS_TIMEOUT) {
    nse->event_done = 1;
    if (FILESPACE_LENGTH(&nse->iobuf) > 0) {
      nse->status = NSE_STATUS_SUCCESS;
    } else {    
      nse->status = NSE_STATUS_TIMEOUT;
    }
  } else if (status == NSE_STATUS_CANCELLED) {
    nse->status = status;
    nse->event_done = 1;
  } else if (status == NSE_STATUS_SUCCESS) {
    while(((buflen = read(nse->iod->sd, buf, sizeof(buf))) > 0) ||
	  ( buflen == -1 && socket_errno() == EINTR)) {
      fscat(&nse->iobuf, buf, buflen);
    }
    if (buflen == 0) {
      nse->event_done = 1;
      nse->eof = 1;
      if (FILESPACE_LENGTH(&nse->iobuf) > 0) {
	nse->status = NSE_STATUS_SUCCESS;
      } else {    
	nse->status = NSE_STATUS_EOF;
      }
    } else if (buflen > 0 || (buflen == -1 && socket_errno() == EAGAIN)) {
      /* We decide whether we have read enough to return */
      
      switch(nse->readinfo.read_type) {
      case NSOCK_READ:
	nse->status = NSE_STATUS_SUCCESS;
	nse->event_done = 1;
	break;
      case NSOCK_READBYTES:
	if (FILESPACE_LENGTH(&nse->iobuf) >= nse->readinfo.num) {
	  nse->status = NSE_STATUS_SUCCESS;
	  nse->event_done = 1;
	} 
	/* else we are not done */
	break;
      case NSOCK_READLINES:
	/* Lets count the number of lines we have ... */
	count = 0;
	len = FILESPACE_LENGTH(&nse->iobuf) -1;
	str = FILESPACE_STR(&nse->iobuf);
	for(count=0; len >= 0; len--) {
	  if (str[len] == '\n') {
	    count++;
	    if ((int) count >= nse->readinfo.num)
	      break;
	  }
	}
	if ((int) count >= nse->readinfo.num) {
	  nse->event_done = 1;
	  nse->status = NSE_STATUS_SUCCESS;
	} 
	/* Else we are not done */
	break;
      default:
	assert(0);
	break; /* unreached */
      }
    } else {  /* read() returned non-EINTR error ... */
      nse->event_done = 1;
      nse->status = NSE_STATUS_ERROR;
      nse->errnum = socket_errno();
    }
  } else {
    assert(0); /* Currently we only know about TIMEOUT and SUCCESS callbacks */
  }
  
  /* If we asked for an event dispatch, we are done reading on the socket so 
     we can take it off the descriptor list ... */
  if (nse->event_done) {
    FD_CLR(nse->iod->sd, &ms->mioi.fds_master_r);
    FD_CLR(nse->iod->sd, &ms->mioi.fds_results_r);
    // Note -- I only want to decrement IOD if there are no other events hinging on it. 
    // For example, there could be a READ and WRITE outstanding at once
    if (nse->iod->events_pending <= 1 && ms->mioi.max_sd == nse->iod->sd)
      ms->mioi.max_sd--;

  }
  
  return;
}

  /* Iterate through all the event lists (such as 
     connect_events, read_events, timer_events, etc) and take action
     for those that have completed (due to timeout, i/o, etc) */

static void iterate_through_event_lists(mspool *nsp) {
    gh_list_elem *current, *next;
    msevent *nse;

    gh_list *event_lists[] = { &nsp->evl.connect_events,
                               &nsp->evl.read_events,
                               &nsp->evl.write_events,
                               &nsp->evl.timer_events,
                               0
                             };
    int current_list_idx;
    nsp->evl.next_ev.tv_sec = 0; /* Clear it -- We will find the next
				   event as we go through the list */

  /* We keep the events seperate because we want to handle them in the 
     order: connect => read => write => timer for several reasons:
     1) Makes sure we have gone through all the net i/o events before 
        a timer expires (would be a shame to timeout after the data was 
        available but before we delivered the events
     2) The connect() results often lead to a read or write that can be 
        processed in the same cycle.  In the same way, read() often 
	leads to write().
  */

    /* foreach list */
    for(current_list_idx = 0; event_lists[current_list_idx] != NULL;
	current_list_idx++) {

    /* foreach element in the list */
      for(current = GH_LIST_FIRST_ELEM(event_lists[current_list_idx]);
	  current != NULL; current = next) {

	nse = (msevent *) GH_LIST_ELEM_DATA(current);
	if ( ! nse->event_done) {	
	  switch(nse->type) {
	  case NSE_TYPE_CONNECT:
	    if (FD_ISSET(nse->iod->sd, &nsp->mioi.fds_results_r) ||
		FD_ISSET(nse->iod->sd, &nsp->mioi.fds_results_w) ||
		FD_ISSET(nse->iod->sd, &nsp->mioi.fds_results_x)) {
	      handle_connect_result(nsp, nse, NSE_STATUS_SUCCESS);
	    } 
	    if (!nse->event_done && nse->timeout.tv_sec &&
		TIMEVAL_NSEC_SUBTRACT(nse->timeout, nsock_tod) <= 0) {
	      handle_connect_result(nsp, nse, NSE_STATUS_TIMEOUT);
	    }
	    break;
	    
	  case NSE_TYPE_READ:
	    if (FD_ISSET(nse->iod->sd, &nsp->mioi.fds_results_r)) {
	      handle_read_result(nsp, nse, NSE_STATUS_SUCCESS);
	    }
	    if (!nse->event_done && nse->timeout.tv_sec &&
		TIMEVAL_NSEC_SUBTRACT(nse->timeout, nsock_tod) <= 0) {
	      handle_read_result(nsp, nse, NSE_STATUS_TIMEOUT);
	    }
	    break;
	    
	  case NSE_TYPE_WRITE:
	    if (FD_ISSET(nse->iod->sd, &nsp->mioi.fds_results_w)) {
	      handle_write_result(nsp, nse, NSE_STATUS_SUCCESS);
	    }
	    if (!nse->event_done && nse->timeout.tv_sec &&
		TIMEVAL_NSEC_SUBTRACT(nse->timeout, nsock_tod) <= 0) {
	      handle_write_result(nsp, nse, NSE_STATUS_TIMEOUT);
	    }
	    break;
	    
	  case NSE_TYPE_TIMER:
	    if (nse->timeout.tv_sec && 
		TIMEVAL_NSEC_SUBTRACT(nse->timeout, nsock_tod) <= 0) {
	      handle_timer_result(nsp, nse, NSE_STATUS_SUCCESS);
	    }
	    break;
	    
	  default:
	    fatal("Event has unknown type (%d)", nse->type);
	    break; /* unreached */
	  }
	}

	if (nse->event_done) {
	  /* WooHoo!  The event is ready to be sent */
	  msevent_dispatch_and_delete(nsp, nse, 1);
	  next = GH_LIST_ELEM_NEXT(current);
	  gh_list_remove_elem(event_lists[current_list_idx], current);
	} else {
	  next = GH_LIST_ELEM_NEXT(current);
	  /* Is this event the next-to-timeout? */
	  if (nse->timeout.tv_sec != 0) {
	    if (nsp->evl.next_ev.tv_sec == 0)
	      nsp->evl.next_ev = nse->timeout;
	    else if (TIMEVAL_NSEC_SUBTRACT(nsp->evl.next_ev, nse->timeout) > 0)
	      nsp->evl.next_ev = nse->timeout;
	  }
	}
      }
    }
}

/* Here is the all important looping function that tells the event
   engine to start up and begin processing events.  It will continue
   until all events have been delivered (including new ones started
   from event handlers), or the msec_timeout is reached, or a major
   error has occured.  Use -1 if you don't want to set a maximum time
   for it to run.  A timeout of 0 will return after 1 non-blocking
   loop.  The nsock loop can be restarted again after it returns.  For
   example you could do a series of 15 second runs, allowing you to do
   other stuff between them */
enum nsock_loopstatus nsock_loop(nsock_pool nsp, int msec_timeout) {
mspool *ms = (mspool *) nsp;
struct timeval loop_timeout;
int msecs_left;
unsigned long loopnum = 0;
enum nsock_loopstatus quitstatus = NSOCK_LOOP_ERROR;

gettimeofday(&nsock_tod, NULL);

 if (msec_timeout > 0) { 
   TIMEVAL_NSEC_ADD(loop_timeout, nsock_tod, msec_timeout);
 } else if (msec_timeout < -1) {
   ms->errnum = EINVAL;
   return NSOCK_LOOP_ERROR;
 }
 msecs_left = msec_timeout;
 
 if (ms->tracelevel > 1) {
   if (msec_timeout >= 0) {
     nsock_trace(ms, "nsock_loop() started (timeout=%dms). %d events pending", msec_timeout, 
		 ms->evl.events_pending);
   } else {
     nsock_trace(ms, "nsock_loop() started (no timeout). %d events pending", ms->evl.events_pending);
   }
 }

while(1) {

  if (ms->evl.events_pending == 0) {
    /* if no events at all are pending, then none can be created until
       we quit nsock_loop() -- so we do that now. */
    quitstatus = NSOCK_LOOP_NOEVENTS;
    break;
  }
 
  if (msec_timeout > 0) {
    msecs_left = MAX(0, TIMEVAL_NSEC_SUBTRACT(loop_timeout, nsock_tod));
    if (msecs_left == 0 && loopnum > 0) {
      quitstatus = NSOCK_LOOP_TIMEOUT;
      break;
    }
  }

  if (wait_for_events(ms, msecs_left) == -1) {  
    quitstatus = NSOCK_LOOP_ERROR;
    break;
  }
  
  /* Now we go through the event lists, doing callbacks for those which
     have completed */
  iterate_through_event_lists(ms);

  gettimeofday(&nsock_tod, NULL); /* we do this at end because there is one
			       at beginning of function */
  loopnum++;
}

 return quitstatus;
}

/* Grab the latest time as recorded by the nsock library, which does
   so at least once per event loop (in main_loop).  Not only does this
   function (generally) avoid a system call, but in many circumstances
   it is better to use nsock's time rather than the system time.  If
   nsock has never obtained the time when you call it, it will do so
   before returning */
const struct timeval *nsock_gettimeofday() {
  if (nsock_tod.tv_sec == 0) 
    gettimeofday(&nsock_tod, NULL);

  return &nsock_tod;
}


/* Adds an event to the appropriate nsp event list, handles housekeeping
   such as adjusting the descriptor select/poll lists, registering the
   timeout value, etc. */
void nsp_add_event(mspool *nsp, msevent *nse) {

  /* First lets do the event-type independant stuff -- starting with
     timeouts */
  if (nse->event_done) { 
    nsp->evl.next_ev = nsock_tod;
  } else {
    if (nse->timeout.tv_sec != 0) {
      if (nsp->evl.next_ev.tv_sec == 0) {
	nsp->evl.next_ev = nse->timeout;
      } else if (TIMEVAL_NSEC_SUBTRACT(nsp->evl.next_ev, nse->timeout) > 0) {
	nsp->evl.next_ev = nse->timeout;
      }
    }
  }

  nsp->evl.events_pending++;

  /* Now we do the event type specific actions */
  switch(nse->type) {
  case NSE_TYPE_CONNECT:
    if (!nse->event_done) {
      FD_SET( nse->iod->sd, &nsp->mioi.fds_master_r);
      FD_SET( nse->iod->sd, &nsp->mioi.fds_master_w);
      FD_SET( nse->iod->sd, &nsp->mioi.fds_master_x);
      nsp->mioi.max_sd = MAX(nsp->mioi.max_sd, nse->iod->sd);
    }
    gh_list_prepend(&nsp->evl.connect_events, nse);
    break;

  case NSE_TYPE_READ:
    if (!nse->event_done) {
      FD_SET( nse->iod->sd, &nsp->mioi.fds_master_r);
      nsp->mioi.max_sd = MAX(nsp->mioi.max_sd, nse->iod->sd);
    }
    gh_list_prepend(&nsp->evl.read_events, nse);
    break;

  case NSE_TYPE_WRITE:
    if (!nse->event_done) {
      FD_SET( nse->iod->sd, &nsp->mioi.fds_master_w);
      nsp->mioi.max_sd = MAX(nsp->mioi.max_sd, nse->iod->sd);
    }
    gh_list_prepend(&nsp->evl.write_events, nse);
    break;

  case NSE_TYPE_TIMER:
    gh_list_prepend(&nsp->evl.timer_events, nse);
    break;

  default:
    assert(0);
    break; /* unreached */

  }
}


void nsock_trace(mspool *ms, char *fmt, ...) {
va_list  ap;
int elapsedTimeMS;

elapsedTimeMS = TIMEVAL_MSEC_SUBTRACT(nsock_tod, ms->tracebasetime);
va_start(ap, fmt);
fflush(stdout);
printf("NSOCK (%.4fs) ", elapsedTimeMS / 1000.0);
vfprintf(stdout, fmt, ap);
printf("\n"); 
va_end(ap);
return;
}

/* An event has been completed and the handler is about to be called.  This function
   writes out tracing data about the event if neccessary */
void nsock_trace_handler_callback(mspool *ms, msevent *nse) {
  msiod *nsi;
  char *str;
  int strlength = 0;
  char displaystr[256];

  if (ms->tracelevel == 0)
    return;

  nsi = nse->iod;

  // Some types have special tracing treatment
  switch(nse->type) {
  case NSE_TYPE_CONNECT:
    nsock_trace(ms, "Callback: %s %s for EID %li [%s:%hi]",  nse_type2str(nse->type), 
		nse_status2str(nse->status), nse->id, inet_ntoa(nsi->peer), nsi->peerport);
    break;

  case NSE_TYPE_READ:
    if (nse->status != NSE_STATUS_SUCCESS)
      nsock_trace(ms, "Callback: %s %s for EID %li", 
		  nse_type2str(nse->type), nse_status2str(nse->status), nse->id);
    else {    
      str = nse_readbuf(nse, &strlength);
      if (ms->tracelevel > 1 && strlength < 80) {
	memcpy(displaystr, ": ", 2);
	memcpy(displaystr + 2, str, strlength);
	displaystr[2 + strlength] = '\0';
	replacenonprintable(displaystr + 2, strlength, '.');
      } else displaystr[0] = '\0';
      nsock_trace(ms, "Callback: %s %s for EID %li %s(%d bytes)%s", 
		  nse_type2str(nse->type), nse_status2str(nse->status), nse->id, 
		  nse_eof(nse)? "[EOF]" : "", strlength, displaystr);
    }
    break;

  case NSE_TYPE_WRITE:
  case NSE_TYPE_TIMER:
    nsock_trace(ms, "Callback: %s %s for EID %li", 
		nse_type2str(nse->type), nse_status2str(nse->status), nse->id);
    break;
  }

}
